/**
 *	@file GradientFile.h
 *	@brief file in header, generated from 'data\gradient400.png'
 *		using BWT / RLE / Huff / BASE64 pipe
 *	@date 2012-10-13 16:30:51
 *	@author FileToHeader (built Dec 10 2013)
 */

#pragma once
#ifndef __DATA_GRADIENT400_PNG_INCLUDED
#define __DATA_GRADIENT400_PNG_INCLUDED

/**
 *	@brief file in header, generated from 'data\gradient400.png'
 */
class CGradientFile {
public:
	/**
	 *	@brief gets size of the data
	 *	@return Returns size of the data, in bytes, not including the terminating zero.
	 */
	static inline size_t n_Size()
	{
		return 535;
	}

	/**
	 *	@brief determines whether the stored string is dynamically allocated
	 *	@return Returns true if the caller of p_Data() needs to free the memory.
	 */
	static inline bool b_Dynamic()
	{
		return true;
	}

	/**
	 *	@brief gets CRC32
	 *	@return Returns CRC32, or zero in case the data was modified (CRC not available).
	 */
	static inline uint32_t n_CRC32()
	{
		return 0x72c899a1;
	}

	/**
	 *	@brief gets data
	 *
	 *	@return Returns data of the file on success, 0 on failure (not enough memory).
	 *
	 *	@note The data (although may be binary), is additionally null terminated to make
	 *		it easier to use (the terminating null is not counted in n_Size()).
	 *	@note See also b_Dynamic().
	 */
	static inline char *p_Data()
	{
		const char *p_s_base64 = "AAAAAAAAAAAAAAAAAQAAAAAAAAAHAAAAKgAAAFoAAABMAAAAAAAAAAAAAAAAAAAAA"
			"AAAAAAAAAAAAAAAAAAAAAAe3AHpEAgqAgQGCyfJYj3VGwUHWp/blsxfI8g2/jyt9tLOjCTKXkreYAzaLBFjl4h"
			"m18bEwxNRvLOBO5N0Ppz/+fcaREXt06rjAwkKpw4PrBS2t7i+GRzAwUDR1CXZ7ikr7y7w8TEyM/I19Dg5+v1IT"
			"lJTVFVbZGVucXJ2eHl+oIKFhoeLkZWYmZujDRIVFhcYHR8hIkkoLS8w5edH6nOA60Z3pKWmfH2piZqNb67zsLF"
			"1bbS1jo9oubr1Q0JBa5J7nfuexYTHZ13LXFnPVtZQ2HpM3Uvf4Sstm4NGEF9hFg1xngwD4JMdd6Rg+Csi4LAyA"
			"N176TIDHnU4wBHSHyZR6nazCghHab9TwlnWfVrSa8EadJJ2ZIANi26esADi3aH8A6AVRQQkaqTKYfxQ7+fjPmZ"
			"c9nEkf1JY16akwD6GwAtl54oW0O48sXuP5pQMamHw7foFFbnVRbGsYshyKs5RU/AKNmOXgGZk7yIAR/oYOmm72"
			"7derpDQTFjRWs+XrWrmGkwWEoUe3qaXExrpZ1IvJeYSpLqzCSGlTQhaWIOKNGDPVKJ0vaxG2533MLQCsXWWo9I"
			"WoOPjoAV2vbFlzZmGS0cvnVc8wOBw7yoaAA4LT3/UzK/m2ccAUHX/vY5U5AHD8LxLqalbxmZ+N67Qlw7XTX389"
			"bZs3VWH5LlPRbSJ+2Ko7jeFkIKR8IkfH4dysgpJv7ENdXweVvm0m1FoOLLTpmK13GQ0R12jc3ar8UvJSQPmBXn"
			"ZmhvWy5ykAFYsUSXiRYuUBd/q9PETkZs8Eaxd+3Pbef65R6JgbG4g6FQmlSwA+eMT3wd9+wpx1WsAcaIv8kQGE"
			"pCDeRGzboXWoOwPUyLNmzRuzitihf/uj58hGGszqiek69IXCePeAJ7kXolKiXO9GRS4qT1siYXmX1NSQD7ThJY"
			"ATpUBfunw";
		return p_Call_Decompress(p_s_base64, 535, 528, 326);
	}

protected:
	/**
	 *	@copydoc CFTHDecompressCode_4d0633d3::p_Decode_BWT_MTF_RLE_Huff_Base64
	 */
	static inline char *p_Call_Decompress(const char *p_s_base64,
		int n_data_size, int n_huff_stage_size, uint32_t n_bwt_primary_index);
};

#ifndef __FILE_TO_HEADER_DECOMPRESS_4D0633D3_INCLUDED
#define __FILE_TO_HEADER_DECOMPRESS_4D0633D3_INCLUDED

/**
 *	@brief decompression routine implementation
 *	@note While this allows all the compressed files to share decompressor implementation,
 *		it restricts all the files to be included in a single .cpp file, otherwise linking
 *		fails because of duplicate symbols (remove body of decompress function in such case).
 */
class CFTHDecompressCode_4d0633d3 {
public:
	/**
	 *	@brief decompression routine
	 *
	 *	@param[in] p_s_base64 is null-terminated string containing BASE-64 encoded compressed data
	 *	@param[in] n_data_size is size of the decompressed data
	 *	@param[in] n_huff_stage_size is size of the compressed data before Huffman encoding
	 *	@param[in] n_bwt_primary_index is primary index for the BWT algorithm
	 *
	 *	@return Returns pointer to the decompressed data on success, 0 on failure.
	 *
	 *	@note The data (although may be binary), is additionally null terminated.
	 */
	static uint8_t *p_Decode_BWT_MTF_RLE_Huff_Base64(const char *p_s_base64,
		int n_data_size, int n_huff_stage_size, uint32_t n_bwt_primary_index)
	{
		int n_base64_length = strlen(p_s_base64);
		int n_base64_out_size = (n_base64_length / 4) * 3 +
			max(int(n_base64_length % 4) - 1, 0);

		int n_prehuff_size = max(n_base64_out_size, n_data_size);
		int n_temp_buffer_size = max(n_prehuff_size + n_huff_stage_size,
			int(n_data_size * (1 + sizeof(uint32_t))));
		uint8_t *p_temp_buffer;
		if(!(p_temp_buffer = new(std::nothrow) uint8_t[n_temp_buffer_size]))
			return 0;;
		uint8_t *p_data;
		if(!(p_data = new(std::nothrow) uint8_t[n_data_size + 1])) {
			delete[] p_temp_buffer;
			return 0;
		}
		p_data[n_data_size] = 0; // !!

		{
			int p_ic_tab[256];
			for(int i = 0; i < 256; ++ i) {
				p_ic_tab[i] = (isalpha(i))? (i | 32) - 'a' + ((i < 'a')? 0 : 26) :
					(isdigit(i))? i - '0' + 52 : (i == '+')? 62 : 63;
			}
			uint8_t *p_src = (uint8_t*)p_s_base64, *p_dst = p_temp_buffer;
			for(const uint8_t *p_end = p_src + n_base64_length - 3; p_src < p_end; p_src += 4, p_dst += 3) {
				uint8_t p_tmp[] = {p_ic_tab[p_src[1]], p_ic_tab[p_src[2]]};
				*p_dst = (p_ic_tab[p_src[0]] << 2) | (p_tmp[0] >> 4);
				p_dst[1] = (p_tmp[0] << 4) | (p_tmp[1] >> 2);
				p_dst[2] = (p_tmp[1] << 6) | p_ic_tab[p_src[3]];
			}
			if(n_base64_length % 4 == 3) {
				uint8_t n_tmp = p_ic_tab[p_src[1]];
				*p_dst = (p_ic_tab[p_src[0]] << 2) | (n_tmp >> 4);
				p_dst[1] = (n_tmp << 4) | ((p_ic_tab[p_src[2]] >> 2) & 0xf); // fixed byte order
			} else if(n_base64_length % 4 == 2)
				*p_dst = (p_ic_tab[p_src[0]] << 2) | (p_ic_tab[p_src[1]] >> 4);
		}
		{
			const uint8_t *p_huff_src = p_temp_buffer, *p_huff_end = p_huff_src + n_base64_out_size;
			_ASSERTE(p_huff_src + 16 * sizeof(int32_t) <= p_huff_end);
			const int32_t *p_code_num = (const int32_t*)p_huff_src; p_huff_src += 16 * sizeof(int32_t);
			int p_min_code[16], p_max_code[16], p_table_off[16], n_code_num = 0;
			for(int i = 0, n_cw = 0; i < 16; ++ i, n_cw <<= 1) {
				p_min_code[i] = n_cw;
				p_max_code[i] = (n_cw += p_code_num[i]);
				p_table_off[i] = n_code_num - p_min_code[i];
				n_code_num += p_code_num[i];
			}
			_ASSERTE(p_huff_src + n_code_num <= p_huff_end);
			const uint8_t *p_symbol = p_huff_src; p_huff_src += n_code_num;
			uint8_t *p_huff_dest = p_temp_buffer + n_prehuff_size;
			for(int n_byte = 0, n_bit_num = 0, n_counter = n_huff_stage_size; n_counter --;) {
				for(int i = 0, n_code = 0;; ++ i) {
					_ASSERTE(i < 16);
					if(!(n_bit_num --)) {
						_ASSERTE(p_huff_src < p_huff_end);
						n_byte = *p_huff_src ++;
						n_bit_num = 7;
					}
					n_code = (n_code << 1) | ((n_byte >> 7) & 1);
					n_byte <<= 1;
					if(n_code >= p_min_code[i] && n_code < p_max_code[i]) {
						*p_huff_dest ++ = p_symbol[n_code + p_table_off[i]];
						break;
					}
				}
			}
		}
		{
			const uint8_t *p_rle_src = p_temp_buffer + n_prehuff_size,
				*p_src_end = p_rle_src + n_huff_stage_size;
			for(uint8_t *p_rle_dest = p_temp_buffer; p_rle_src != p_src_end;) {
				uint8_t n_code = *p_rle_src ++;
				bool b_compressed = n_code & 1;
				int n_run_length = (n_code >> 1) + 1;
				_ASSERTE(p_rle_src + ((b_compressed)? 1 : n_run_length) <= p_src_end);
				while(n_run_length --)
					*p_rle_dest ++ = (b_compressed)? *p_rle_src : *p_rle_src ++;
				if(b_compressed)
					++ p_rle_src;
			}
		}
		{
			uint8_t p_buckets[256];
			for(int i = 0; i < 256; ++ i)
				p_buckets[i] = (uint8_t)i;
			const uint8_t *p_src = p_temp_buffer, *p_end = p_temp_buffer + n_data_size;
			for(; p_src != p_end; ++ p_src) {
				uint8_t *p_dest = (uint8_t*)p_src, n_index = *p_src;
				*p_dest = p_buckets[n_index];
				if(n_index) {
					for(int i = n_index; i > 0; -- i)
						p_buckets[i] = p_buckets[i - 1];
					p_buckets[0] = *p_dest;
				}
			}
		}
		{
			uint32_t p_buckets[256] = {0}, n_sum = 0;
			uint32_t *p_index_list = (uint32_t*)(p_temp_buffer + n_data_size);
			const uint8_t *p_src = p_temp_buffer, *p_end = p_temp_buffer + n_data_size;
			for(uint32_t *p_dest_idx = p_index_list; p_src != p_end; ++ p_src, ++ p_dest_idx)
				*p_dest_idx = (++ p_buckets[*p_src]) - 1;
 			for(int i = 0; i < 256; ++ i) {
				uint32_t n_value = p_buckets[i];
				p_buckets[i] = n_sum;
				n_sum += n_value;
			}
			uint8_t *p_dest = p_data + n_data_size - 1, *p_begin_1 = p_data - 1;
			for(uint32_t i = n_bwt_primary_index; p_dest != p_begin_1; -- p_dest,
			   i = p_index_list[i] + p_buckets[p_temp_buffer[i]])
				*p_dest = p_temp_buffer[i];
		}

		delete[] p_temp_buffer;
		
		return p_data;
	}
};

#endif // __FILE_TO_HEADER_DECOMPRESS_4D0633D3_INCLUDED

inline char *CGradientFile::p_Call_Decompress(const char *p_s_base64,
	int n_data_size, int n_huff_stage_size, uint32_t n_bwt_primary_index)
{
	return (char*)CFTHDecompressCode_4d0633d3::p_Decode_BWT_MTF_RLE_Huff_Base64(p_s_base64,
		n_data_size, n_huff_stage_size, n_bwt_primary_index);
}
// a dummy function that calls the decompressor, which is now declared
// (allows us to keep it at the end of the file)

#endif // __DATA_GRADIENT400_PNG_INCLUDED
