/*
								+--------------------------------+
								|                                |
								|      ***  HTTP latch  ***      |
								|                                |
								|  Copyright  -tHE SWINe- 2013  |
								|                                |
								|            Main.cpp            |
								|                                |
								+--------------------------------+
*/

#if (defined(_WIN32) || defined(_WIN64)) && defined(_DEBUG)
#include <crtdbg.h>
//#undef _ASSERTE
#endif // (_WIN32 || _WIN64) && _DEBUG
// win32 memory debugger

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <vector>
#include <string>
#include <map>
#include <memory> // auto_ptr
#include "../UberLame_src/Socket.h"
#include "../UberLame_src/MinMax.h"
#include "../UberLame_src/Integer.h"
#include "../UberLame_src/Hash.h"
#include "../UberLame_src/proto/HTTP_Client.h"
#include "../UberLame_src/proto/HTTP_Server.h"
#include "../UberLame_src/Timer.h"
#include "../UberLame_src/Dir.h" // only PRIsizeB

#if defined(_WIN32) || defined(_WIN64)
#define NOMINMAX
#include <windows.h>
#include <wincrypt.h>

class CWinCrypt {
protected:
	HCRYPTPROV m_h_crypt;

public:
	CWinCrypt()
	{
		CryptAcquireContext(&m_h_crypt, 0, 0, PROV_RSA_FULL, CRYPT_SILENT);
	}

	~CWinCrypt()
	{
		CryptReleaseContext(m_h_crypt, 0);
	}

	HCRYPTPROV h_Crypt()
	{
		return m_h_crypt;
	}
} _win_crypt; // win32 random number generator initializer

#else // _WIN32 || _WIN64
//#include <random.h> // unix random generator (don't have?)
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h> // sleep()
static void Sleep(int n_millis)
{
	struct timespec t_time;
	memset(&t_time, 0, sizeof(t_time));
	t_time.tv_sec = n_millis / 1000;
	t_time.tv_nsec = (n_millis % 1000) * 1000000;
	nanosleep(&t_time, NULL);
}
#endif // _WIN32 || _WIN64

namespace app_data {

#include "FaviconFile.h"
#include "GradientFile.h"
#include "IndexFile.h"
#include "JSSHALicenseFile.h"
#include "ScriptsFile.h"
#include "StyleSheetFile.h"
// data files

} // ~app_data

#if defined(_WIN32) || defined(_WIN64)
#define strncasecmp(a,b,n) _strnicmp((a), (b), (n))
#define strcasecmp(a,b) _stricmp((a), (b))
#endif //_WIN32 || _WIN64

#if (defined(_WIN32) || defined(_WIN64)) && defined(_DEBUG)

#ifndef __USE_MALLOC
#pragma message("warning: memory debugger enabled and __USE_MALLOC not defined: STL will leak\n")
#endif // !__USE_MALLOC

#ifdef __XML_NODE_STATIC_DUMMY
#pragma message("warning: memory debugger enabled and __XML_NODE_STATIC_DUMMY defined: XML will cause two memory leaks\n")
#endif // __XML_NODE_STATIC_DUMMY

class CDumpMemoryLeaks {
public:
	~CDumpMemoryLeaks()
	{
		fprintf(stderr, "\nDEBUG: output from _CrtDumpMemoryLeaks():\n");
		_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_FILE);
		_CrtSetReportFile(_CRT_WARN, _CRTDBG_FILE_STDERR);
		_CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_FILE);
		_CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDERR);
		_CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
		_CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
		_CrtDumpMemoryLeaks();
		// win32 memory debugger (doesn't quite work with new stl unless __USE_MALLOC is defined)
	}
};

static CDumpMemoryLeaks memory_debugger;

#endif // (_WIN32 || _WIN64) && _DEBUG

#define __LATCH_VERSION__ "1.00"

/*
 *	const char *p_s_IpString(const TNetAddress &r_t_address)
 *		- returns ip address in format n.n.n.n:port
 *		- note returned string is stored in static array and
 *		  therefore it's contents may change with every call
 *		  to this function
 */
const char *p_s_IpString(const TNetAddress &r_t_address)
{
	static char p_s_ip[64];
	unsigned int n_ip;
	memcpy(&n_ip, &r_t_address.t_Address().sin_addr, sizeof(int));
	sprintf(p_s_ip, "%d.%d.%d.%d:%d", n_ip & 0xff, (n_ip >> 8) & 0xff,
		(n_ip >> 16) & 0xff, n_ip >> 24, htons(r_t_address.t_Address().sin_port));
	return p_s_ip;

	// todo - support ipv6
}

static const char *p_s_html_head =
	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
	"<html>\n"
	"<head>\n"
	"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
	"<style type=\"text/css\">\n"
	"body {\n"
	"	font-family: tahoma, verdana, sans-serif;\n"
	"}\n"
	"h2 {\n"
	"	margin: 0px 50px;\n"
	"	padding: 0px 16px;\n"
	"	border-bottom-style: solid;\n"
	"	border-bottom-width: 1px;\n"
	"	border-bottom-color: black;\n"
	"}\n"
	"a {\n"
	"	text-decoration: none;\n"
	"}\n"
	"a:hover {\n"
	"	text-decoration: underline;\n"
	"}\n"
	"div {\n"
	"	font-size: 10px;\n"
	"	text-align: right;\n"
	"	margin: 0px 50px;\n"
	"	padding: 0px 16px;\n"
	"	border-top-style: solid;\n"
	"	border-top-width: 1px;\n"
	"	border-top-color: black;\n"
	"}\n"
	"</style>\n"
	"<title>HTTP Latch</title>\n"
	"</head>\n"
	"<body>\n";
// HTML head (with <body>)

static const char *p_s_html_tail =
	"<div>&Uuml;berLame tiny HTTP server v" __LATCH_VERSION__
	" (build " __DATE__ ") :: made by -tHE SWINe-</div>\n"
	"</body>\n"
	"</html>\n";
// HTML tail (with </body>)

class CMyHTTP_Server : public CHTTP_Server {
protected:
	bool m_b_verbose;

	enum {
		buffer_Size = 65536
	};

	struct TAddressFilter {
		uint32_t n_address;
		uint32_t n_mask;

		inline TAddressFilter()
		{}

		inline TAddressFilter(uint32_t _n_address, uint32_t _n_mask)
			:n_address(_n_address & _n_mask), n_mask(_n_mask)
		{}

		inline TAddressFilter(uint8_t n_addr_0, uint8_t n_addr_1,
			uint8_t n_addr_2, uint8_t n_addr_3, size_t n_sig_bits)
			:n_address((n_addr_0 << 24) | (n_addr_1 << 16) | (n_addr_2 << 8) | n_addr_3),
			n_mask((uint32_t(1) << (n_sig_bits - 1)) | ((uint32_t(1) << (n_sig_bits - 1)) - 1))
		{
			n_mask <<= 32 - n_sig_bits; // !!
			n_address &= n_mask;
		}

		bool b_Match(const TNetAddress &r_t_addr) const
		{
			uint32_t n_addr = *(const uint32_t*)&r_t_addr.t_Address().sin_addr;
			n_addr = htonl(n_addr);

			return (n_addr & n_mask) == n_address;
		}
	};

	std::vector<TAddressFilter> m_blacklist, m_whitelist;
	std::string m_s_password;
	const char *m_p_s_logfile;
	const char *m_p_s_latch_message;
	bool m_b_suicide;
	CSemaphore &m_r_shutdown_sema;

	static char n_Unicode_to_US_English(unsigned int n_char)
	{
		return (n_char < 128)? n_char : '_'; // don't use a question mark here
	}

public:
	CMyHTTP_Server(const std::string &r_s_password, CSemaphore &r_shutdown_sema,
		bool b_verbose = false, const char *p_s_latch_message = 0, const char *p_s_logfile = 0)
		:m_b_verbose(b_verbose), m_s_password(r_s_password),
		m_p_s_logfile(p_s_logfile), m_p_s_latch_message(p_s_latch_message),
		m_b_suicide(false), m_r_shutdown_sema(r_shutdown_sema)
	{}

	virtual bool Handle_BadRequest(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		fprintf(stderr, "%s - warning: %s %s from %s ended up as 400: bad request\n",
			p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
			p_s_IpString(t_address));
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 400: bad request\n",
					p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
					p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: "
				"Bad Request (400)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_BadRequest(p_header, p_socket);
		}
		return CHTTP_Server::Handle_BadRequest(p_header, p_socket, s_response.c_str());
	}

	virtual bool Handle_NotFound(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		fprintf(stderr, "%s - warning: %s %s from %s ended up as 404: not found\n",
			p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
			p_s_IpString(t_address));
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 404: not found\n",
					p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
					p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: "
				"File Not Found (404)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_NotFound(p_header, p_socket);
		}
		return CHTTP_Server::Handle_NotFound(p_header, p_socket, s_response.c_str());
	}

	virtual bool Handle_CantBeOpened(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		fprintf(stderr, "%s - warning: %s %s from %s ended up as 404: can't be opened\n",
			p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
			p_s_IpString(t_address));
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 404: can't be opened\n",
					p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
					p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: File can not "
				"be accessed by the server process (404)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_NotFound(p_header, p_socket);
		}
		return CHTTP_Server::Handle_NotFound(p_header, p_socket, s_response.c_str());
	}

	virtual bool Handle_Forbidden(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 403: forbidden\n",
					p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
					p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: "
				"Access to this file is forbidden (403)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_NotFound(p_header, p_socket);
		}
		return CHTTP_Server::Handle_NotFound(p_header, p_socket, s_response.c_str());
	}

	virtual bool Handle_InternalServerError(
		const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		fprintf(stderr, "%s - warning: %s %s from %s ended up as 500: internal server error\n",
			p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
			p_s_IpString(t_address));
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 500: "
					"internal server error\n", p_s_DateString(s_time), p_header->p_s_Method(),
					p_header->p_s_Request_URI(), p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: "
				"Internal Server Error (500)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_InternalServerError(p_header, p_socket);
		}
		return CHTTP_Server::Handle_InternalServerError(p_header, p_socket, s_response.c_str());
	}

	virtual bool Handle_NotImplemented(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) const
	{
		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		std::string s_time;
		fprintf(stderr, "%s - warning: %s %s from %s ended up as 501: not implemented\n",
			p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
			p_s_IpString(t_address));
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "%s - warning: %s %s from %s ended up as 501: not implemented\n",
					p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
					p_s_IpString(t_address));
				fclose(p_log);
			}
		}
		// add log entry

		std::string s_response;
		__STL_UT_TRY {
			s_response += p_s_html_head;
			s_response += "<h2 style=\"border-bottom-style: none;\">Error: "
				"Not Implemented (501)</h2>\n";
			s_response += p_s_html_tail;
		} __STL_UT_CATCH_BAD_ALLOC {
			return CHTTP_Server::Handle_NotImplemented(p_header, p_socket);
		}
		return CHTTP_Server::Handle_NotImplemented(p_header, p_socket, s_response.c_str());
	}

	class CLameAuth {
	protected:
		double m_f_tslice_expiration_time;
		CTimer m_timer;

		struct TTimeSlice {
			double f_expiration_time;
			uint32_t n_time_slice;

			inline TTimeSlice()
			{}

			inline TTimeSlice(double _f_expiration_time, uint32_t _n_time_slice)
				:f_expiration_time(_f_expiration_time), n_time_slice(_n_time_slice)
			{}
		};

		std::map<std::string, TTimeSlice> m_tslice_map;

	public:
		CLameAuth(double f_tslice_expiration_time = 5 * 60)
			:m_f_tslice_expiration_time(f_tslice_expiration_time)
		{}

		bool b_IsLameAuth(const char *p_s_filename) const
		{
			return strstr(p_s_filename, "_lauth") != 0; // if it begins with _lauth, it is lame auth
		}

		static int n_Rotate16(int i, int n) // rotates a 16-bit number
		{
			return ((i << n) | (i >> n)) % 65536;
		}

		template <class THash>
		static THash t_HMAC(const void *p_key, size_t n_key_size,
			const void *p_message, size_t n_message_size) // note this currently only works with SHA1, endianness 
		{
			const char *p_s_key = (const char*)p_key;
			size_t n_key_size_bits = n_key_size * 8;
			// this is used as key for HMAC

			//printf("hmac key is %s (%d bits)\n", p_s_key, n_key_size_bits); // debug

			// t_odo - wrap HMAC to a function
			// todo - test examples on wiki

			typedef int TBlockSizeCheck[!(THash::block_Size_Bits % 32)];
			typedef int THashSizeCheck[!(THash::hash_Size_Bits % 32)];
			// make sure these are multiple of 32 (not done a nice way, but it will work)

			uint32_t keyToUse[THash::block_Size_Bits / 32] = {0}; // 64 bytes for MD5 / SHA1
			if(n_key_size > THash::block_Size_Bits / 8) {
				CStreamHash<THash> hash;
				hash.Process_Data(p_s_key, n_key_size_bits / 8);
				THash t_key = hash.t_Result();
				if(THash::word_Is_BigEndian) {
					for(int i = 0; i < THash::hash_Size_Bits / 32; ++ i)
						keyToUse[i] = htonl(t_key[i]); // note this is untested, especially the use of htonl
				} else {
					memcpy(keyToUse, &t_key[0], THash::hash_Size_Bits / 8);
					// just copy, no endian conversion
				}
				// zero-padded hash as a key
			} else if(n_key_size <= THash::block_Size_Bits / 8) {
#if 0
				for(int i = 0; i < n_key_size_bits; i += 8)
					keyToUse[i >> 5] |= uint32_t(uint8_t(p_s_key[i / 8])) << (i % 32); // this doesn't solve endianness
#else
				memcpy(keyToUse, p_s_key, n_key_size_bits / 8); // this doesn't solve endianness
#endif
				// convert the key to 4 x 16 bytes, one byte at a time
			}
			// create block key

			uint32_t p_i_key_pad[THash::block_Size_Bits / 32],
				p_o_key_pad[THash::block_Size_Bits / 32];
			for(int i = 0; i < THash::block_Size_Bits / 32; ++ i) {
				p_i_key_pad[i] = keyToUse[i] ^ 0x36363636U;
				p_o_key_pad[i] = keyToUse[i] ^ 0x5c5c5c5cU;
			}
			// calculate pads

			THash t_hmac;
			{
				THash t_intermed_hash;
				{
					CStreamHash<THash> hash;
					hash.Process_Data(p_i_key_pad, THash::block_Size_Bits / 8); // p_i_key_pad
#if 0
					// this doesn't solve hash size, etc. this was written for TSHA1
					std::vector<uint32_t> str_to_hash;
					if(!stl_ut::Resize_To_N(str_to_hash, (s_orig_request_uri.length() + 3) / 4, 0))
						return false;
					for(size_t i = 0, n = s_orig_request_uri.length() * sizeof(char) * 8; i < n; i += 8)
						str_to_hash[i >> 5] |= uint32_t(uint8_t(s_orig_request_uri[i / 8])) << (i % 32); // this doesn't solve endianness
					hash.Process_Data(&str_to_hash[0], s_orig_request_uri.length() * sizeof(char)); // message
#else
					hash.Process_Data(p_message, n_message_size); // message // this doesn't solve endianness
#endif
					t_intermed_hash = hash.t_Result();
					if(THash::word_Is_BigEndian) {
						for(int i = 0; i < THash::hash_Size_Bits / 32; ++ i)
							t_intermed_hash[i] = htonl(t_intermed_hash[i]); // !!
					}
				}
				CStreamHash<THash> hash;
				hash.Process_Data(p_o_key_pad, THash::block_Size_Bits / 8); // p_o_key_pad
				hash.Process_Data(&t_intermed_hash[0], THash::hash_Size_Bits / 8); // t_intermed_hash
				t_hmac = hash.t_Result();
			}
			// calculate HMAC

			return t_hmac;
		}

		static TSHA1 t_SaltPassword(const void *p_password, size_t n_password_size,
			uint32_t n_time_slice, int n_round_num = 997)
		{
			char p_s_time_slice_id[256];
			sprintf(p_s_time_slice_id, "&time-slice-id=%u", n_time_slice);
			// get time slice as string

			TSHA1 t_salted_pwd;
			{
				CStreamHash<TSHA1> hash;
				hash.Process_Data(p_password, n_password_size);
				hash.Process_Data(p_s_time_slice_id, strlen(p_s_time_slice_id) * sizeof(char));
				t_salted_pwd = hash.t_Result();

				//printf("password + salt is %s%s\n", s_password.c_str(), s_time_slice_id.c_str()); // debug
				//printf("hash(password + salt) is %s\n", t_salted_pwd.p_s_ToString()); // debug

				for(int i = 0; i < 997; ++ i) { // t_odo - make this a prime
					const char *p_s_hash = t_salted_pwd.p_s_ToString();
					_ASSERTE(strlen(p_s_hash) == 40);
					hash.Reset();
					hash.Process_Data(p_s_hash, 40 * sizeof(char));
					char p_s_rot[64];
					sprintf(p_s_rot, "%u", n_Rotate16(i, 10));
					hash.Process_Data(p_s_rot, strlen(p_s_rot) * sizeof(char));
					t_salted_pwd = hash.t_Result();
				}
			}
			// calculate salted password, this is used as key for HMAC

			return t_salted_pwd;
		}

		struct TURI_Context {
			const CHTTP_Header *p_header;
			CHTTP_Socket *p_socket;

			TNetAddress t_address;
			std::string s_address;

			bool b_method_get;
			std::string s_str_uri;
			std::string s_file;
			bool b_uri_has_arguments;
			std::string s_orig_request_uri;
			mutable std::map<std::string, std::string> arg_map; // [] requires no-const

			inline TURI_Context()
			{}

			inline TURI_Context(const CHTTP_Header *_p_header, CHTTP_Socket *_p_socket) // throws(std::bad_alloc)
				:p_header(_p_header), p_socket(_p_socket)
			{
				b_method_get = strcmp(p_header->p_s_Method(), "HEAD") != 0;
				_ASSERTE(!b_method_get || !strcmp(p_header->p_s_Method(), "GET") ||
					!strcmp(p_header->p_s_Method(), "POST"));
				// determine wheter GET/POST or HEAD

				if(!CURI_Utils::Unescape_URI_UTF8(s_str_uri, p_header->s_Request_URI(), n_Unicode_to_US_English))
					throw std::bad_alloc();
				if(!CURI_Utils::Get_FileName_Part(s_file, s_str_uri, "index.html"))
					throw std::bad_alloc();
				// extract desired filename along with path; handle absolute URI's

				p_socket->GetPeerName(t_address);
				s_address = p_s_IpString(t_address);
				size_t n_pos;
				if((n_pos = s_address.find(':')) != std::string::npos)
					s_address.erase(n_pos);
				// get peer name (for time slices, sessions, ...)

				b_uri_has_arguments = s_str_uri.find('?') != std::string::npos;
				if(!CURI_Utils::Split_URI_Params(arg_map, s_str_uri, false, false))
					throw std::bad_alloc();
				// parse args

				if(!CURI_Utils::Remove_URI_Param(s_orig_request_uri, s_str_uri, "hmac", true))
					throw std::bad_alloc();
				// get request uri without hmac information (if any args follow, they are contained)
			}
		};

		bool Hanlde_LameAuth(bool &r_b_valid_request, std::string &s_response, const TURI_Context &r_req_context) // t_odo - remove the buffer, add response as std::string reference
		{
			r_b_valid_request = true;
			// just assume ...

			// todo - split this to funcions, don't repeat code, don't make a mess
			// todo - use user logged in state, don't let the account be used even if not logged in
			// (anyone can send messages containing scores and screw the scores for the user)
			// todo - add request id counter per user to make hashes different without forcing too low
			// secret time
			// todo - it is actually pretty easy, just maintain the counter at client side,
			// make it a part of request. the server side just makes sure that user increments
			// the counter (requests with the same value of counter will be ignored)
			// todo - add mutexes arround lameauth
			// todo - catch uncaught exceptions

			if(!r_req_context.s_file.compare("_lauth/time-slice.xml")) {
				if(r_req_context.b_uri_has_arguments) {
					r_b_valid_request = false;
					return true;
				}
				// no '?'s allowed

				double f_time = m_timer.f_Time();
				try {
					if((m_tslice_map.find(r_req_context.s_address) == m_tslice_map.end()) ||
					   m_tslice_map[r_req_context.s_address].f_expiration_time < f_time) {
						m_tslice_map[r_req_context.s_address] = TTimeSlice(f_time + m_f_tslice_expiration_time,
							n_CSRandom()); // nice, only called limited times / time interval, no DOS attack here
					}
				} catch(std::bad_alloc&) {
					return false;
				}
				// makes sure the time slice does exist, and is valid

				const TTimeSlice &t_time_slice_id = m_tslice_map[r_req_context.s_address];
				if(!stl_ut::Format(s_response, "<?xml version=\"1.0\" encoding=\"utf-8\" "
				   "standalone=\"yes\"?><time-slice-id valid_msec=\"%u\">%u</time-slice-id>",
				   (uint32_t)((t_time_slice_id.f_expiration_time - m_timer.f_Time()) * 1000),
				   t_time_slice_id.n_time_slice))
					return false;
				// create a response xml for the time slice
			} else {
				r_b_valid_request = false;
				return true;
				// unknown lauth request
			}

			return true;
		}

		bool b_LameAuth_VerifyAuthentication(bool &r_b_valid_request,
			const TURI_Context &r_req_context, const std::string &s_password)
		{
			r_b_valid_request = true;

			if(r_req_context.arg_map.find("hmac") == r_req_context.arg_map.end()) {
				r_b_valid_request = false;
				return true;
			}
			const std::string &r_s_hmac = r_req_context.arg_map["hmac"];
			// get username and HMAC

			//const std::string &s_password = r_s_password; // already arg
			// read password from db ...

			double f_time = m_timer.f_Time();
			if((m_tslice_map.find(r_req_context.s_address) == m_tslice_map.end()) || // note that the time slice may not exist!
			   m_tslice_map[r_req_context.s_address].f_expiration_time + 1 < f_time) {
				r_b_valid_request = false; // time slice expired, no access
				return true;
			}
			const TTimeSlice &t_time_slice_id = m_tslice_map[r_req_context.s_address];
			// read time slice from the list, give it some reserve before expiring

			TSHA1 t_salted_pwd = t_SaltPassword(s_password.data(),
				s_password.length() * sizeof(char), t_time_slice_id.n_time_slice); // t_odo - see if the time slice is allocated! might be producing random numbers here
			// calculate salted password, this is used as key for HMAC

			_ASSERTE(strlen(t_salted_pwd.p_s_ToString()) == 40);
			TSHA1 t_hmac = t_HMAC<TSHA1>(t_salted_pwd.p_s_ToString(), 40 * sizeof(char),
				r_req_context.s_orig_request_uri.data(),
				r_req_context.s_orig_request_uri.length() * sizeof(char));
			// calculate HMAC

			const char *p_s_hmac = t_hmac.p_s_ToString();
			//printf("calculated expected HMAC: %s\n", p_s_hmac); // debug
			// this is how is the HMAC supposed to look like

			r_b_valid_request = !strcasecmp(p_s_hmac, r_s_hmac.c_str());
			// in case HMAC is valid, access is granted

			return true;
		}

		static uint32_t n_CSRandom()
		{
#if defined(_WIN32) || defined(_WIN64)
			uint32_t n;
			CryptGenRandom(_win_crypt.h_Crypt(), sizeof(n), (BYTE*)&n);
			return n;
#else // _WIN32 || _WIN64
#if 0
			uint32_t n;
			get_random_bytes(&n, sizeof(n));
			return n;
#elif 1
			static const char *p_s_randpath = "/dev/random";
 			int fp = open(p_s_randpath, O_RDONLY);
			if(fp < 0){
				fprintf(stderr, "Could not open -%s- for reading.\n", p_s_randpath);
				exit(-1);
			}
			// open as read only

			uint32_t n;
			if(read(fp, &n, sizeof(uint32_t)) < 0){
				fprintf(stderr, "Could not read from -%s-.\n", p_s_randpath);
				close(fp);
				exit(-1);
			}
			// get random data

			close(fp);	
			return n;
#else // 0
			return rand(); // t_odo - replace by a secure random number generator
#endif // 0
#endif // _WIN32 || _WIN64
		}
	};

	CLameAuth m_lame_auth;

	static bool b_IsDummyResource(const char *p_s_file)
	{
		return!strcmp(p_s_file, "index.html") || !strcmp(p_s_file, "scripts.js") ||
			!strcmp(p_s_file, "gradient400.png") || !strcmp(p_s_file, "style.css") ||
			!strcmp(p_s_file, "favicon.ico") || !strcmp(p_s_file, "jsSHA_LICENSE.txt");
	}

	bool Serve_File(const CHTTP_Header *p_header, CHTTP_Socket *p_socket) // throws(std::bad_alloc)
	{
		_ASSERTE(p_header && p_header->p_s_Request_URI());
		// there must be request URI

		if(!strstr(p_header->p_s_Status_Line(), "HTTP/1.0") && !p_header->p_s_Field("Host"))
			return Handle_BadRequest(p_header, p_socket);
		// header must contain host field (aparts from HTTP/1.0 requests)

		CLameAuth::TURI_Context t_context(p_header, p_socket); // throws!

		bool b_method_get = t_context.b_method_get;
		// determine wheter GET/POST or HEAD

		std::string &str_uri = t_context.s_str_uri;
		std::string &s_file = t_context.s_file;
		const char *p_s_file = s_file.c_str();
		// extract desired filename along with path; handle absolute URI's

		std::string s_lauth_response;
		bool b_delete_buffer = true;

		uint64_t n_file_size = 0;
		unsigned char *p_file_buffer = 0;
		const char *p_s_content_type = "application/octet-stream";
		const char *p_s_content_disposition = "inline";
		bool b_chunked_file = false;
		{
			bool b_directory;
			b_directory = !strcmp(p_s_file, ".") || (strlen(p_s_file) &&
				p_s_file[strlen(p_s_file) - 1] == '/');
			// decide wheter URL points to a directory

			if(b_directory)
				return Handle_Forbidden(p_header, p_socket); // directory listing forbidden
			if(!m_lame_auth.b_IsLameAuth(p_s_file)) {
				//TFileInfo t_file(p_s_file); // no filesystem access
				if(s_file.compare("trigger.xml") && !b_IsDummyResource(p_s_file))
					return Handle_NotFound(p_header, p_socket); // not found
			}
			// handle cases when referenced file is in fact directory

			if(b_directory) // a dir
				return Handle_Forbidden(p_header, p_socket); // directory listing forbidden
			else { // a file
				// todo - see if the file is accessible
				// todo - see if the file requires authentication

				p_s_content_type = p_s_MimeType(p_s_file);
				p_s_content_disposition = (strncmp(p_s_content_type,
					"application/", 12))? "inline" : "attachment";
				// get file mime type (want to display images / documents directly in browser window)

				if(m_lame_auth.b_IsLameAuth(p_s_file)) {
					bool b_valid_request;
					if(!m_lame_auth.Hanlde_LameAuth(b_valid_request, s_lauth_response, t_context))
						return Handle_InternalServerError(p_header, p_socket);
					// handle lame auth resources (such as time-slice.xml, login.xml, possibly more)

					if(!b_valid_request && s_lauth_response.empty())
						return Handle_BadRequest(p_header, p_socket);
					// in case lauth didn't provide any response and the request is invalid,
					// show bad request, rather than internal server error

					p_file_buffer = (unsigned char*)s_lauth_response.data();
					n_file_size = s_lauth_response.length();
					b_delete_buffer = false; // !!
				} else {
					// this is a physical file

					if(!s_file.compare("trigger.xml")) {
						_ASSERTE(t_context.arg_map.find("username") != t_context.arg_map.end());
						const std::string &r_s_user = t_context.arg_map["username"];
						// get username

						bool b_verify;
						if(!m_lame_auth.b_LameAuth_VerifyAuthentication(b_verify, t_context, m_s_password))
							return Handle_InternalServerError(p_header, p_socket);
						// in case the file requires authentication, verify it. if it fails, show forbidden

						if(b_verify) {
							if(m_b_verbose)
								fprintf(stderr, "debug: latch triggered\n");
							if(m_p_s_logfile) {
								FILE *p_log;
								if(p_log = fopen(m_p_s_logfile, "a")) {
									fprintf(p_log, "debug: latch triggered\n");
									fclose(p_log);
								}
							}
							// log

							p_file_buffer = (uint8_t*)"<?xml version=\"1.0\" encoding=\"utf-8\" "
								"standalone=\"yes\"?><result value=\"1\" reason=\"Latch unlocked\"/>";
							// form the reply

							m_b_suicide = true;
							// do kill the server
						} else {
							if(m_b_verbose)
								fprintf(stderr, "debug: latch NOT triggered: authentication failed\n");
							if(m_p_s_logfile) {
								FILE *p_log;
								if(p_log = fopen(m_p_s_logfile, "a")) {
									fprintf(p_log, "debug: latch NOT triggered: authentication failed\n");
									fclose(p_log);
								}
							}
							// log

							p_file_buffer = (uint8_t*)"<?xml version=\"1.0\" encoding=\"utf-8\" "
								"standalone=\"yes\"?><result value=\"0\" reason=\"Latch NOT unlocked:"
								" authentication failed\"/>";
							// form the reply
						}
						n_file_size = strlen((char*)p_file_buffer) * sizeof(char);
						b_delete_buffer = false; // !!
						// simple answer
					} else {
						if(!strcmp(p_s_file, "index.html")) {
							p_file_buffer = (uint8_t*)app_data::CIndexFile::p_Data();

							try {
								const char *p_s_message_para = "<p id=\"message-para\"></p>";
								const size_t n_message_para_length = strlen(p_s_message_para);
								// find the message para

								std::string s_index_html = (const char*)p_file_buffer;
								size_t n_mp_pos = s_index_html.find(p_s_message_para);
								_ASSERTE(n_mp_pos != std::string::npos);

								if(!m_p_s_latch_message) {
									s_index_html.erase(n_mp_pos, n_message_para_length);
									// erase the para

									if(app_data::CIndexFile::b_Dynamic()) {
										_ASSERTE((s_index_html.length() + 1) * sizeof(char) < app_data::CIndexFile::n_Size());
										// we just deleted stuff, it should be smaller

										memcpy(p_file_buffer, s_index_html.c_str(), (s_index_html.length() + 1) * sizeof(char));
									} else
										p_file_buffer = (uint8_t*)strdup(s_index_html.c_str()); // checked for null afterwards
									// copy the data
								} else {
									n_mp_pos += n_message_para_length - strlen("</p>");
									// seek to the point where to put the data

									s_index_html.insert(n_mp_pos, m_p_s_latch_message);
									// add the message inside the para

									if(app_data::CIndexFile::b_Dynamic())
										delete[] p_file_buffer;
									p_file_buffer = (uint8_t*)strdup(s_index_html.c_str()); // checked for null afterwards
									// put it there
								}
							} catch(std::bad_alloc&) {
								return Handle_InternalServerError(p_header, p_socket);
							}
							n_file_size = strlen((const char*)p_file_buffer) * sizeof(char);
							b_delete_buffer = true;
						} else if(!strcmp(p_s_file, "scripts.js")) {
							p_file_buffer = (uint8_t*)app_data::CScriptsFile::p_Data();
							n_file_size = app_data::CScriptsFile::n_Size();
							b_delete_buffer = app_data::CScriptsFile::b_Dynamic();
						} else if(!strcmp(p_s_file, "gradient400.png")) {
							p_file_buffer = (uint8_t*)app_data::CGradientFile::p_Data();
							n_file_size = app_data::CGradientFile::n_Size();
							b_delete_buffer = app_data::CGradientFile::b_Dynamic();
						} else if(!strcmp(p_s_file, "style.css")) {
							p_file_buffer = (uint8_t*)app_data::CStyleSheetFile::p_Data();
							n_file_size = app_data::CStyleSheetFile::n_Size();
							b_delete_buffer = app_data::CStyleSheetFile::b_Dynamic();
						} else if(!strcmp(p_s_file, "favicon.ico")) {
							p_file_buffer = (uint8_t*)app_data::CFaviconFile::p_Data();
							n_file_size = app_data::CFaviconFile::n_Size();
							b_delete_buffer = app_data::CFaviconFile::b_Dynamic();
						} else if(!strcmp(p_s_file, "jsSHA_LICENSE.txt")) {
							p_file_buffer = (uint8_t*)app_data::CJSSHALicenseFile::p_Data();
							n_file_size = app_data::CJSSHALicenseFile::n_Size();
							b_delete_buffer = app_data::CJSSHALicenseFile::b_Dynamic();
						} else
							return Handle_CantBeOpened(p_header, p_socket);
						if(!p_file_buffer)
							return Handle_InternalServerError(p_header, p_socket);

						/*TFileInfo file(p_s_file); // no filesystem access
						if(!file.b_Valid() || !file.b_exists)
							return Handle_CantBeOpened(p_header, p_socket);
						n_file_size = file.n_Size64();
						// get file size

						b_chunked_file = true;
						if(b_method_get && !(p_file_buffer = new(std::nothrow)
						   unsigned char[size_t(min(uint64_t(buffer_Size), n_file_size))]))
							return Handle_InternalServerError(p_header, p_socket);
						// allocate data buffer*/
					}
				}
			}

			_ASSERTE(!b_directory);
			// no directories beyond this point
		}

		std::auto_ptr<unsigned char> buffer_ptr((b_method_get && b_delete_buffer)? p_file_buffer : 0);
		// if HEAD, no buffer is allocated (also, it might be allocated in a std::string and doesn't need deletion)

		std::string s_response, s_date;
		if(!stl_ut::Format(s_response, "HTTP/1.1 200 OK\r\n"
		   "Date: %s\r\n"
		   "Content-Length: " PRIu64 "\r\n"
		   "Content-Type: %s\r\n"
		   "Content-Disposition: %s\r\n"
		   "Connection: close\r\n\r\n", p_s_DateString(s_date), n_file_size,
		   p_s_content_type, p_s_content_disposition))
			return Handle_InternalServerError(p_header, p_socket);
		// create response string (contains no html body)

		CTimer timer;
		bool b_printed_cr = false;
		for(int n_pass = 0; n_pass < ((b_method_get)? 2 : 1); ++ n_pass) {
			if(n_pass && b_chunked_file) {
				// access to filesystem forbidden
#if 0
				FILE *p_fr;
				if(!(p_fr = fopen(p_s_file, "rb")))
					return Handle_NotFound(p_header, p_socket);
				// open file ...

				uint64_t n_next_log = 1048576;
				for(uint64_t n_total_written = 0; n_total_written < n_file_size;) {
					if(n_total_written >= n_next_log) {
						b_printed_cr = true;
						n_next_log = n_total_written + 1048576;
						std::string s_time;
						fprintf(stderr, "\r%s - uploaded " PRIsizeB "B, out of " PRIsizeB "B @ "
							PRIsizeB "B/s", p_s_DateString(s_time), PRIsizeBparams(n_total_written),
							PRIsizeBparams(n_file_size), PRIsizeBparams(n_total_written /
							timer.f_Time()));
					}
					// see how much it downloaded

					int n_remains = int(min(uint64_t(buffer_Size), n_file_size - n_total_written));
					if(fread(p_file_buffer, n_remains, 1, p_fr) != 1) {
						if(b_printed_cr)
							printf("\n");
						fclose(p_fr);
						return false;
					}
					// read chunk of data

					unsigned int n_to_write = n_remains;
					const unsigned char *p_write_ptr = p_file_buffer;
					while(n_to_write) {
						int n_written;
						if((n_written = p_socket->n_Write(p_write_ptr, n_to_write)) <= 0) {
							if(b_printed_cr)
								printf("\n");
							fclose(p_fr);
							return false;
						}
						n_to_write -= n_written;
						p_write_ptr += n_written;
					}
					// write chunk

					n_total_written += n_remains;
				}

				fclose(p_fr);
#endif // 0
			} else {
				_ASSERTE(n_file_size <= SIZE_MAX); // just generated files
				size_t n_to_write = (n_pass)? size_t(n_file_size) : s_response.length();
				const unsigned char *p_write_ptr = (n_pass)? p_file_buffer :
					(const unsigned char*)s_response.data();
				while(n_to_write) {
					int n_written;
					if((n_written = p_socket->n_Write(p_write_ptr, n_to_write)) <= 0)
						return false;
					n_to_write -= n_written;
					p_write_ptr += n_written;
				}
			}
		}
		// send data

		if(b_printed_cr)
			printf("\n");

		return true;
	}

	virtual bool Handle_HEAD(const CHTTP_Header *p_header, CHTTP_Socket *p_socket, int n_tid)
	{
		return Handle_GET(p_header, p_socket, n_tid);
	}

	virtual bool Handle_POST(const CHTTP_Header *p_header, CHTTP_Socket *p_socket, int n_tid)
	{
		return Handle_GET(p_header, p_socket, n_tid);
	}

	bool Add_BlacklistAddr(uint8_t n_addr_0, uint8_t n_addr_1, uint8_t n_addr_2,
		uint8_t n_addr_3, size_t n_sig_bits)
	{
		return stl_ut::Resize_Add_1More(m_blacklist,
			TAddressFilter(n_addr_0, n_addr_1, n_addr_2, n_addr_3, n_sig_bits));
	}

	bool Add_WhitelistAddr(uint8_t n_addr_0, uint8_t n_addr_1, uint8_t n_addr_2,
		uint8_t n_addr_3, size_t n_sig_bits)
	{
		return stl_ut::Resize_Add_1More(m_whitelist,
			TAddressFilter(n_addr_0, n_addr_1, n_addr_2, n_addr_3, n_sig_bits));
	}

	bool b_IsAllowed_Address(const TNetAddress &r_t_addr) const
	{
		for(size_t i = 0, n = m_blacklist.size(); i < n; ++ i) {
			if(m_blacklist[i].b_Match(r_t_addr))
				return false;
		}
		// match addresses using blacklist

		if(m_whitelist.empty())
			return true;
		// in case whitelist is not used, the address may pass now

		for(size_t i = 0, n = m_whitelist.size(); i < n; ++ i) {
			if(m_whitelist[i].b_Match(r_t_addr))
				return true;
		}
		// match addresses using whitelist

		return false;
	}

	virtual bool Handle_GET(const CHTTP_Header *p_header, CHTTP_Socket *p_socket, int n_tid)
	{
		int n_pid = n_tid;

		if(m_b_verbose)
			fprintf(stderr, "--- worker with pid %d ---\n", n_pid);
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "--- worker with pid %d ---\n", n_pid);
				fclose(p_log);
			}
		}
		// add log entry

		TNetAddress t_address;
		p_socket->GetPeerName(t_address);
		// get peer name

		if(!b_IsAllowed_Address(t_address)) {
			// handle addresses that do not pass the ip filter

			std::string s_time;
			if(m_b_verbose) {
				fprintf(stderr, "%s - incoming connection from %s; request %s %s "
					"(ignored, access denied)\n", p_s_DateString(s_time),
					p_s_IpString(t_address), p_header->p_s_Method(), p_header->p_s_Request_URI());
			}
			if(m_p_s_logfile) {
				FILE *p_log;
				if(p_log = fopen(m_p_s_logfile, "a")) {
					fprintf(p_log, "%s - incoming connection from %s; request %s %s (ignored,"
						" access denied)\n", p_s_DateString(s_time), p_s_IpString(t_address),
						p_header->p_s_Method(), p_header->p_s_Request_URI());
					fclose(p_log);
				}
			}
			// add log entry
		} else {
			if(m_b_verbose) {
				std::string s_time;
				if(m_b_verbose) {
					fprintf(stderr, "%s - incoming connection from %s; request %s %s\n",
						p_s_DateString(s_time), p_s_IpString(t_address), p_header->p_s_Method(),
						p_header->p_s_Request_URI());
				}
				if(m_p_s_logfile) {
					FILE *p_log;
					if(p_log = fopen(m_p_s_logfile, "a")) {
						fprintf(p_log, "%s - incoming connection from %s; request %s %s\n",
							p_s_DateString(s_time), p_s_IpString(t_address),
							p_header->p_s_Method(), p_header->p_s_Request_URI());
						fclose(p_log);
					}
				}
				// add log entry
			}

			try {
				if(!Serve_File(p_header, p_socket)) {
					TNetAddress t_address;
					p_socket->GetPeerName(t_address);
					std::string s_time;
					fprintf(stderr, "%s - error: error completing %s %s request from %s ...\n",
						p_s_DateString(s_time), p_header->p_s_Method(), p_header->p_s_Request_URI(),
						p_s_IpString(t_address));
					if(m_b_verbose)
						fprintf(stderr, "--- ~worker with pid %d ---\n", n_pid);
					if(m_p_s_logfile) {
						FILE *p_log;
						if(p_log = fopen(m_p_s_logfile, "a")) {
							fprintf(p_log, "%s - error: error completing %s %s request from %s ...\n",
								p_s_DateString(s_time), p_header->p_s_Method(),
								p_header->p_s_Request_URI(), p_s_IpString(t_address));
							fprintf(p_log, "--- ~worker with pid %d ---\n", n_pid);
							fclose(p_log);
						}
					}
					// add log entry
					return false;
				}
			} catch(std::bad_alloc&) {
				Handle_InternalServerError(p_header, p_socket);
			}
		}
		// perform simple IP filtering

		if(m_b_verbose)
			fprintf(stderr, "--- ~worker with pid %d ---\n", n_pid);
		if(m_p_s_logfile) {
			FILE *p_log;
			if(p_log = fopen(m_p_s_logfile, "a")) {
				fprintf(p_log, "--- ~worker with pid %d ---\n", n_pid);
				fclose(p_log);
			}
		}
		// add log entry

		if(m_b_suicide)
			m_r_shutdown_sema.Signal();
		// kill the server, if decided that the latch is opened

		return true;
	}

	virtual bool Handle_PUT(const CHTTP_Header *p_header, CHTTP_Socket *p_socket, int n_tid) const
	{
		return CHTTP_Server::Handle_NotImplemented(p_header, p_socket);
	}

protected:
	static const char *p_s_MimeType(const char *p_s_filename)
	{
		const char *p_s_ext = strrchr(p_s_filename, '.');
		if(!p_s_ext)
			p_s_ext = "";
		else
			++ p_s_ext;
		// get file extension

		const struct {
			const char *p_s_ext;
			const char *p_s_mime;
		} p_type_table[] = {
			{"jpg", "image/jpeg"},
			{"gif", "image/gif"},
			{"png", "image/png"},
			{"ico", "image/icon"},

			{"c", "text/plain"},
			{"cpp", "text/plain"},
			{"h", "text/plain"},
			{"hpp", "text/plain"},
			{"txt", "text/plain"},

			{"htm", "text/html"},
			{"html", "text/html"},
			{"xml", "text/xml"},
			{"css", "text/css"},
			{"js", "text/javascript"},
		};
		static const int n_type_num = sizeof(p_type_table) / sizeof(p_type_table[0]);
		// content-type table

		for(int i = 0; i < n_type_num; ++ i) {
			if(!strcasecmp(p_s_ext, p_type_table[i].p_s_ext))
				return p_type_table[i].p_s_mime;
		}
		// get file mime type (want to display images / documents directly in browser window)

		return "application/octet-stream";
	}
};

/*
 *								=== ~CMyHTTP_Server ===
 */

bool GetHostName(std::string &r_s_hostname)
{
	char *p_s_hostname;
	size_t n_length = 1;
	for(int i = 0; i < 16; ++ i) {
		if(!(p_s_hostname = new(std::nothrow) char[n_length + 1]))
			return false;
		memset(p_s_hostname, 0, (n_length + 1) * sizeof(char));
		if(!gethostname(p_s_hostname, n_length) && p_s_hostname[n_length] == 0) {
			r_s_hostname.erase();
			r_s_hostname.reserve(n_length);
			if(r_s_hostname.capacity() < n_length) {
				delete[] p_s_hostname;
				return false;
			}
			r_s_hostname.assign(p_s_hostname);
			delete[] p_s_hostname;
			return true;
		}
		n_length = (n_length * 2) | 1;
		delete[] p_s_hostname;
	}
	return false;
}

int main(int n_arg_num, const char **p_arg_list)
{
	{
		const char *p_s_passfile = 0, *p_s_logfile = 0;
		int n_port = 80;
		int n_listen = 25;
		bool b_verbose = false;
		std::vector<const char*> whitelist;
		std::vector<const char*> blacklist;
		const char *p_s_latch_message = 0;

		bool b_show_help = false;
		for(int i = 1; i < n_arg_num; ++ i) {
			if(!strcmp(p_arg_list[i], "-h") || !strcmp(p_arg_list[i], "--help"))
				b_show_help = true;
			else if(!strcmp(p_arg_list[i], "-v") || !strcmp(p_arg_list[i], "--verbose"))
				b_verbose = true;
			else if(!strcmp(p_arg_list[i], "-q") || !strcmp(p_arg_list[i], "--quiet"))
				b_verbose = false;
			else if(i + 1 >= n_arg_num) {
				fprintf(stderr, "error: argument \'%s\': unknown argument or needs a value\n");
				return -1;
			} else if(!strcmp(p_arg_list[i], "-w") || !strcmp(p_arg_list[i], "--whitelist")) {
				if(!stl_ut::Resize_Add_1More(whitelist, p_arg_list[++ i])) {
					fprintf(stderr, "error: not enough memory\n");
					return -1;
				}
			} else if(!strcmp(p_arg_list[i], "-b") || !strcmp(p_arg_list[i], "--blacklist")) {
				if(!stl_ut::Resize_Add_1More(blacklist, p_arg_list[++ i])) {
					fprintf(stderr, "error: not enough memory\n");
					return -1;
				}
			} else if(!strcmp(p_arg_list[i], "-p") || !strcmp(p_arg_list[i], "--port"))
				n_port = atol(p_arg_list[++ i]);
			else if(!strcmp(p_arg_list[i], "-l") || !strcmp(p_arg_list[i], "--listen"))
				n_listen = atol(p_arg_list[++ i]);
			else if(!strcmp(p_arg_list[i], "-lf") || !strcmp(p_arg_list[i], "--log"))
				p_s_logfile = p_arg_list[++ i];
			else if(!strcmp(p_arg_list[i], "-pf") || !strcmp(p_arg_list[i], "--password-file"))
				p_s_passfile = p_arg_list[++ i];
			else if(!strcmp(p_arg_list[i], "-m") || !strcmp(p_arg_list[i], "--message"))
				p_s_latch_message = p_arg_list[++ i];
			else
				fprintf(stderr, "error: unknown parameter: %s\n", p_arg_list[i]);
		}
		// parse additional args

		if(!p_s_passfile || b_show_help) {
			printf("synopsis:\n"
				"http_latch -pf|--password-file passFile [[-p|--port] <port>] [[-l|--listen] <listen-count>] "
				"[-v|--verbose] [-h|--help] [-w|--whitelist ip/mask-bits] "
				"[-b|--blacklist ip/mask-bits] [-lf|--log logFile] [-m|--message messageHTML]\n\n"
				"where port is listening port, listen-count is listening queue size\n"
				"-v or --verbose enables verbose output, including download progress\n"
				"-g or --log <filename> enables logging to a file (just adds)\n"
				"-w or --whitelist adds an IP address to whitelist and\n"
				"-b or --blacklist adds an IP address to blacklist\n"
				"\t(use addresses in e.g. 127.0.0.1/32 format)\n"
				"-pf or --password-file sets file with the password\n"
				"-lf or --log sets file with the server log (only appended)\n"
				"-m or --message sets message on latch page (default blank)\n"
				"-h or --help shows this helpscreen\n\n");
		}
		// show help (avoids showing it multiple times)

		if(!p_s_passfile) {
			fprintf(stderr, "error: no password file entered\n");
			return -1;
		}

		if(b_verbose) {
			fprintf(stderr, "HTTP latch " __LATCH_VERSION__ " build %s\n", __DATE__);
			fprintf(stderr, "listen(port=%d, backlog=%d)\n", n_port, n_listen);
			std::string s_hostname;
			const char *p_s_ip = "0.0.0.0";
			if(!GetHostName(s_hostname))
				fprintf(stderr, "error: failed to determine hostname\n");
			else {
				TNetAddress t_addr(s_hostname.c_str(), n_port);
				// create address from hostname

				p_s_ip = p_s_IpString(t_addr);
				// create IP string

				struct hostent *p_host;
				if(p_host = gethostbyaddr((const char*)&t_addr.r_Address().sin_addr, 4, AF_INET)) {
					if(stl_ut::Format(s_hostname, (n_port == 80)?
					   "http://%s/" : "http://%s:%d/", p_host->h_name, n_port))
						p_s_ip = 0; // don't need ip, have full name
				}
				// attempt to get full hostname
			}
			fprintf(stderr, (p_s_ip)? "computer-name=%s (http://%s)\n" :
				"computer-name=%s\n", s_hostname.c_str(), p_s_ip);
		}

		std::string s_password;
		{
			FILE *p_fr;
			if(!(p_fr = fopen(p_s_passfile, "r"))) {
				fprintf(stderr, "error: failed to open \'%s\'\n", p_s_passfile);
				return -1;
			}
			if(!stl_ut::ReadLine(s_password, p_fr)) {
				fclose(p_fr);
				fprintf(stderr, "error: failed to read \'%s\'\n", p_s_passfile);
				return -1;
			}
			fclose(p_fr);
			stl_ut::TrimSpace(s_password);
		}
		// read the password

		if(s_password.empty())
			fprintf(stderr, "error: using empty password\n");
		else if(s_password.length() < 5)
			fprintf(stderr, "error: using short password (%d characters only)\n", int(s_password.length()));
		// check the password

		CSemaphore end_sema(0);
		CMyHTTP_Server server(s_password, end_sema, b_verbose, p_s_latch_message, p_s_logfile);

		for(size_t i = 0, n = whitelist.size(); i < n; ++ i) {
			const char *p_s_ip_port = whitelist[i];
			int p_ip[4], n_mask;
			if(sscanf(p_s_ip_port, "%d.%d.%d.%d/%d", &p_ip[0],
			   &p_ip[1], &p_ip[2], &p_ip[3], &n_mask) != 5)  {
				fprintf(stderr, "error: bad whitelist address: \'%s\'"
					" (ignored; use n.n.n.n/mask format)\n", p_s_ip_port);
				continue;
			}
			if(p_ip[0] == 127 && p_ip[1] == 0 && p_ip[2] == 0 && p_ip[3] == 1 && n_mask == 32)
				printf("whitelisting loopback\n");
			else
				printf("whitelisting \'%s\'\n", p_s_ip_port);
			if(p_ip[0] == 192 && p_ip[1] == 168 && p_ip[2] == 0 && n_mask >= 24)
				printf("whitelisting ICS network (\'%s\')\n", p_s_ip_port);
			else
				printf("whitelisting \'%s\'\n", p_s_ip_port);

			server.Add_WhitelistAddr(p_ip[0], p_ip[1], p_ip[2], p_ip[3], n_mask);
		}
		for(size_t i = 0, n = blacklist.size(); i < n; ++ i) {
			const char *p_s_ip_port = blacklist[i];
			int p_ip[4], n_mask;
			if(sscanf(p_s_ip_port, "%d.%d.%d.%d/%d", &p_ip[0],
			   &p_ip[1], &p_ip[2], &p_ip[3], &n_mask) != 5)  {
				fprintf(stderr, "error: bad blacklist address: \'%s\'"
					" (ignored; use n.n.n.n/mask format)\n", p_s_ip_port);
				continue;
			}

			if(p_ip[0] == 127 && p_ip[1] == 0 && p_ip[2] == 0 && p_ip[3] == 1 && n_mask == 32)
				printf("blacklisting loopback\n");
			else
				printf("blacklisting \'%s\'\n", p_s_ip_port);
			if(p_ip[0] == 192 && p_ip[1] == 168 && p_ip[2] == 0 && n_mask >= 24)
				printf("blacklisting ICS network (\'%s\')\n", p_s_ip_port);
			else
				printf("blacklisting \'%s\'\n", p_s_ip_port);

			server.Add_BlacklistAddr(p_ip[0], p_ip[1], p_ip[2], p_ip[3], n_mask);
		}
		// whitelist some addresses (t_odo - use commandline)

		if(!server.Listen_Start(n_port, n_listen)) { // threaded
			fprintf(stderr, "error: internal server error while CHTTP_Server::Listen_Start()\n");
			return -1;
		}
		// listen (blocking)

		if(!end_sema.Wait()) {
			fprintf(stderr, "fatal error: failed to wait for semaphore\n");
			return -1;
		}
		// wait for server to finish

		Sleep(1000); // wait for server to do stuff, send answers, ...
		if(!server.Listen_End()) {
			fprintf(stderr, "error: internal server error while CHTTP_Server::Listen_End()\n");
			exit(-1);
		}
		// close the server

		if(b_verbose)
			printf("finished ...\n");
		// just that so we know
	}

	return 0;
}
